@@ -3,7 +3,7 @@ typedload
 Module to load data into typed data structures
 """
 
-# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2022 Salvo "LtWorf" Tomaselli
 #
 # typedload is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,12 +22,15 @@ Module to load data into typed data structures
 
 
 import datetime
+from enum import Enum
 import ipaddress
+from functools import reduce
 from pathlib import Path
 from typing import *
 
 from .exceptions import *
 from .typechecks import *
+from .typechecks import discriminatorliterals
 from .helpers import tname
 
 
@@ -226,6 +229,8 @@ class Loader:
 
         self._indexcache = {}  # type: Dict[Type, int]
 
+        self._unionload_discriminatorcache = {}  # type: Dict[Type, Tuple[Optional[str], Optional[Dict[Any, Type]]]]
+
     def index(self, type_: Type[T]) -> int:
         """
         Returns the index in the handlers list
@@ -355,10 +360,22 @@ def _listload(l: Loader, value, type_) -> List:
         raise TypedloadTypeError('Unable to load dictionary as a list', value=value, type_=type_)
     t = type_.__args__[0]
     try:
+        f = l.handlers[l.index(t)][1]
+    except ValueError:
+        raise TypedloadTypeError(
+            'Cannot deal with value of type %s' % tname(type_),
+            value=value,
+            type_=type_
+        )
+    try:
+        # Hopeful load calling the handler directly, skipping load()
+        return [f(l, v, t) for v in value]
+    except TypedloadException as e:
+        # FIXME Once 3.8 is the lowest supported version, I can use := to export the index where the
+        # error happened, removing the need to redo the entire loading of the list in the slow way
+        # A nested error happened, reload everything with load() so we get the detailed exception with path
         return [l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value)]
     except TypeError as e:
-        if isinstance(e, TypedloadException):
-            raise
         raise TypedloadTypeError(str(e), value=value, type_=type_)
 
 
@@ -387,7 +404,23 @@ def _setload(l: Loader, value, type_) -> Set:
     if isinstance(value, dict):
         raise TypedloadTypeError('Unable to load dictionary as a set', value=value, type_=type_)
     t = type_.__args__[0]
-    return {l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value)}
+    try:
+        f = l.handlers[l.index(t)][1]
+    except ValueError:
+        raise TypedloadTypeError(
+            'Cannot deal with value of type %s' % tname(type_),
+            value=value,
+            type_=type_
+        )
+    try:
+        # Hopeful load
+        return {f(l, v, t) for v in value}
+    except TypedloadException:
+        # A failure, reload with annotations
+        # See _listload for more details
+        return {l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value)}
+    except TypeError as e:
+        raise TypedloadTypeError(str(e), value=value, type_=type_)
 
 
 def _frozensetload(l: Loader, value, type_) -> FrozenSet:
@@ -397,7 +430,22 @@ def _frozensetload(l: Loader, value, type_) -> FrozenSet:
     if isinstance(value, dict):
         raise TypedloadTypeError('Unable to load dictionary as a frozenset', value=value, type_=type_)
     t = type_.__args__[0]
-    return frozenset(l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
+    try:
+        f = l.handlers[l.index(t)][1]
+    except ValueError:
+        raise TypedloadTypeError(
+            'Cannot deal with value of type %s' % tname(type_),
+            value=value,
+            type_=type_
+        )
+    try:
+        return frozenset(f(l, v, t) for v in value)
+    except TypedloadException:
+        # A failure, reload with annotations
+        # See _listload for more details
+        return frozenset(l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
+    except TypeError as e:
+        raise TypedloadTypeError(str(e), value=value, type_=type_)
 
 
 def _tupleload(l: Loader, value, type_) -> Tuple:
@@ -412,7 +460,23 @@ def _tupleload(l: Loader, value, type_) -> Tuple:
         args = type_.__tuple_params__
 
     if len(args) == 2 and args[1] == ...: # Tuple[something, ...]
-        return tuple(l.load(v, args[0], annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
+        t = args[0]
+        try:
+            f = l.handlers[l.index(t)][1]
+        except ValueError:
+            raise TypedloadTypeError(
+                'Cannot deal with value of type %s' % tname(type_),
+                value=value,
+                type_=type_
+            )
+        try:
+            return tuple(f(l, v, t) for v in value)
+        except TypedloadException:
+            # A failure, reload with annotations
+            # See _listload for more details
+            return tuple(l.load(v, t, annotation=Annotation(AnnotationType.INDEX, i)) for i, v in enumerate(value))
+        except TypeError as e:
+            raise TypedloadTypeError(str(e), value=value, type_=type_)
     else: # Tuple[something, something, somethingelse]
         if l.failonextra and len(value) > len(args):
             raise TypedloadValueError('Value is too long for type %s' % tname(type_), value=value, type_=type_)
@@ -577,13 +641,19 @@ def _typeddictload(l: Loader, value: Dict[str, Any], type_) -> Any:
 
     if hasattr(type_, '__required_keys__') and hasattr(type_, '__optional_keys__'):
         # TypedDict, since 3.9
-        necessary_fields = type_.__required_keys__
+        necessary_fields = set(type_.__required_keys__)
     elif getattr(type_, '__total__', True) == False:
         # TypedDict, only for 3.8
         necessary_fields = set()
     else:
         necessary_fields = fields
 
+    # Resolve the NotRequired stuff
+    for k, v in type_hints.items():
+        if is_notrequired(v):
+            type_hints[k] = notrequiredtype(v)
+            necessary_fields.discard(k)
+
     return _objloader(l, fields, necessary_fields, type_hints, value, type_)
 
 
@@ -611,6 +681,39 @@ def _unionload(l: Loader, value, type_) -> Any:
     sorted_args = list(args)  # type: List[Type]
     sorted_args.sort(key=lambda i: i in l.basictypes)
 
+    # For object types, bump up the type whose literal is matching
+    if hasattr(value, 'get'):
+        # Seems we have an object
+        # Bump up if the Literal field matches
+        discriminatorscache = l._unionload_discriminatorcache.get(type_)  # type: Optional[Tuple[Optional[str], Optional[Dict[Any, Type]]]]
+
+        # First time generate the deep inspection for literal
+        if discriminatorscache is None:
+            # type → {key: valueset}
+            data = {t: discriminatorliterals(t) for t in args}
+            # shared keys that have literals in every object of the union
+            keys = reduce(lambda a, b: a.intersection(b), (set(v.keys()) for v in data.values()))  # type: Set[str]
+
+            cachedict = {}
+            if keys:
+                key = keys.pop()
+                for t, d in data.items():
+                    for literal in d[key]:
+                        cachedict[literal] = t
+                discriminatorscache = key, cachedict
+            else:
+                discriminatorscache = None, None
+            l._unionload_discriminatorcache[type_] = discriminatorscache
+
+        # Cache is created, use it
+        # It's a tuple key, {value: type}
+        if discriminatorscache[1]:
+            preferredtype = discriminatorscache[1].get(value.get(discriminatorscache[0]))
+            if preferredtype:
+                # Place best value on top
+                sorted_args.remove(preferredtype)
+                sorted_args.insert(0, preferredtype)
+
     # Try all types
     loaded_count = 0
     r = None
